from contextlib import suppress

from .event_handler import EventHandler


class EventHandlerFactoryBase:
    def build(self, event_handler, *, keys=None):
        analyzer = EventHandlerAnalyzer(event_handler, keys)
        event_handler = self._build_from_callable(analyzer)
        event_handler.start()
        return event_handler

    def _build_from_callable(self, analyzer):
        server_connector = self._build_server_connector()
        server_connector.subscribe(analyzer.keys)
        event_handler = EventHandler(server_connector)
        event_handler.handle_event = analyzer.handle_event
        with suppress(AttributeError):
            event_handler.cleanup = analyzer.cleanup
        return event_handler

    def _build_server_connector(self):
        raise NotImplementedError()


class EventHandlerAnalyzer:
    def __init__(self, event_handler, supplied_keys):
        self._event_handler = event_handler
        self._supplied_keys = supplied_keys

    @property
    def keys(self):
        if self._supplied_keys is None:
            try:
                return self._event_handler.keys
            except AttributeError:
                raise ValueError('No keys supplied!')
        return self._supplied_keys

    @property
    def handle_event(self):
        try:
            return self._event_handler.handle_event
        except AttributeError:
            if callable(self._event_handler):
                return self._event_handler
            raise ValueError('Object must implement handle_event or be a callable!')

    @property
    def cleanup(self):
        return self._event_handler.cleanup
